#include <stdio.h>
#include <fcntl.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

/* Use this to allow/disallow network syslog support */
#define NETSYSLOG 1

#ifdef NETSYSLOG
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <resolv.h>
#endif


#define	LOGFILE		"/var/log/syslog"
#define	CONFFILE	"/etc/config/config"

static int	syslog_sfd = -1;
static char	*syslog_ident = (char *) NULL;
static FILE	*syslog_fp = (FILE *) NULL;


void syslog(int prio, const char *form, ...)
{
	va_list ptr;
	char	buf[128], *sp;

	if (syslog_fp == NULL)
		return;

	sp = &buf[0];
	sp += sprintf(sp, "%s: ", syslog_ident);
	va_start(ptr, form);
	sp += vsprintf(sp, form, ptr);
	va_end(ptr);
	sp += sprintf(sp, "\n");

	fprintf(syslog_fp, buf);
	fflush(syslog_fp);

#ifdef NETSYSLOG
	if (syslog_sfd >= 0)
		send(syslog_sfd, buf, (sp - buf), 0);
#endif
}


#ifdef NETSYSLOG

static void opennetlog(void)
{
	struct hostent		*hp;
	struct sockaddr_in	sin;
	char			*whitespace = " 	";
	FILE			*fp;
	char			line[80], *sp;
	char			*ipstr = NULL, *portstr = NULL;

	/* Find syslog machine address and port */
	if ((fp = fopen(CONFFILE, "r")) == (FILE *) NULL)
		return;

	while (fgets(line, sizeof(line), fp) != NULL) {
		if (strncmp(line, "syslog", 6) == 0) {
			sp = &line[6];
			strsep(&sp, whitespace);
			ipstr = strsep(&sp, whitespace);
			portstr = strsep(&sp, whitespace);
			break;
		}
	}
	fclose(fp);
	
	if (ipstr == NULL)
		return;
	if (portstr == NULL)
		portstr = "514";

	sin.sin_family = AF_INET;
	sin.sin_port = htons(atoi(portstr));
	sin.sin_addr.s_addr = inet_addr(ipstr);
	if (sin.sin_addr.s_addr == -1) {
		if ((hp = gethostbyname(ipstr))) {
			sin.sin_family = hp->h_addrtype;
			memcpy(&sin.sin_addr, hp->h_addr, hp->h_length);
		}
	}


	/* Open socket to remote machine for sysloging */
	if ((syslog_sfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		return;
	}

	if (connect(syslog_sfd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		close(syslog_sfd);
		syslog_sfd = -1;
	}
}

#endif /* NETSYSLOG */


void openlog(const char *ident, int i, int j)
{
	syslog_fp = fopen(LOGFILE, "a");
	if ((syslog_ident = (char *) ident) == (char *) NULL) {
		if ((syslog_ident = malloc(16)) != (char *) NULL)
			sprintf(syslog_ident, "[%d]", getpid());
		else
			syslog_ident = "[UKNOWN]";
	}
#ifdef NETSYSLOG
	opennetlog();
#endif
}


void closelog(void)
{
	/* Don't close the log... */
#if 0
	if (syslog_fp != (FILE *) NULL)
		fclose(syslog_fp);
	if (syslog_sfd >= 0)
		close(syslog_sfd_;
	syslog_fp = (FILE *) NULL;
	syslog_sfd = -1;
#endif
}


int setlogmask(int i)
{
	/* Should really implement this :-) */
}

