/*****************************************************************************/

/*
 *	crt0_ram.S -- startup code for MCF5307 ColdFire based NETtel.
 *
 *	(C) Copyright 1999, Greg Ungerer.
 *
 * 	1999/02/24 Modified for the 5307 processor David W. Miller
 *
 *	Modified for the Isicad Prisma 700 (68020, no MMU) by George Harvey
 *
 *	Modified for the Cisco 2500, (68ec030 @ 20MHz) by Koen De Vleeschauwer 
 */

/*****************************************************************************/

#include "linux/autoconf.h"

/*****************************************************************************/

/*
 *
 * The interrupt vectors are located at the start of (at least) 2M of DRAM
 * located at address 0. The kernel loads into RAM starting at address 0x1000. 
 *
 */

#define	MEM_BASE	0x00001000	/* Memory base at address 0x001000 */
/*#define	MEM_END		0x003fffff	/* Memory size 4Mb */
#define	MEM_END		0x007fffff	/* Memory size 4Mb */
#define	VBR_BASE	0		/* Vector address */

/*****************************************************************************/

.global	_start
.global _rambase
.global _ramvec
.global __ramvec
.global	_ramstart
.global	__ramstart
.global	_ramend
.global __ramend

/*****************************************************************************/

.data

/*
 *	Set up the usable of RAM stuff. Size of RAM is determined then
 *	an initial stack set up at the end.
 */
_rambase:
.long	0
_ramvec:
__ramvec:
.long	0
_ramstart:
__ramstart:
.long	0
_ramend:
__ramend:
.long	0

/*****************************************************************************/

.text

/*
 *	This is the codes first entry point. This is where it all
 *	begins...
 */
_start:	
	nop					/* Filler */
	move.w	#0x2700, %sr			/* No interrupts */

	/*
	 * Setup VBR here, otherwise buserror remap will not work.
	 * if dBug was active before (on my SBC with dBug 1.1 of Dec 16 1996)
	 *
	 * bkr@cut.de 19990306
	 *
	 * Note: this is because dBUG points VBR to ROM, making vectors read
	 * only, so the bus trap can't be changed. (RS)
	 */
	move.l	#VBR_BASE, %a7			/* Note VBR can't be read */
	movec   %a7, %VBR
	move.l	%a7, _ramvec			/* Set up vector addr */
	move.l	#MEM_BASE, %a7
	move.l	%a7, _rambase			/* Set up base RAM addr */

	/*
	 *	Determine size of RAM, then set up initial stack.
	 * XXX Fixme
	 * determine ram size automatically
	 */
	/*
	 * Initialize all stack pointers, erring on the cautious side.
	 */
	move.l	#MEM_END, %d0
	move.c	%d0, %isp			/* Set up interrupt stack */
	sub.l	#0x1000, %d0
        move.c  %d0, %msp			/* kernel stack */
        move.l  %d0, _ramend                    /* Set end ram addr */ 
	sub.l	#0x1000, %d0
        move.c  %d0, %usp			/* Just being cautious */

	/*
	 *	Enable CPU internal cache.
	 */
	move.l	#0x00000008, %d0		/* invalidate whole cache */
	movec	%d0,%CACR
	nop

	/* Enable cache */
	move.l	#0x00000001, %d0
	movec	%d0,%CACR
	nop

	/*
	 *	No need to move ROM filesystem above bss :-)
	 *	Just initialize ramstart.
	 */
	lea.l	_ebss, %a1			/* start of ram = end of bss  */
	mov.l	%a1, %d0
	and.l	#0xfffffff0, %d0		/* Align */
	add.l	#0x1e, %d0			
	move.l	%d0, _ramstart			/* Set start of ram */

	/*
	 *	Zero out the bss region.
	 */
	lea.l	_sbss, %a0			/* Get start of bss */
	lea.l	_ebss, %a1			/* Get end of bss */
	clr.l	%d0				/* Set value */
_clear_bss:
	move.l	%d0, (%a0)+			/* Clear each word */
	cmp.l	%a0, %a1			/* Check if at end */
	bge	_clear_bss			/* Just in case ebss=sbss */

	/*
	 *	Assember start up done, start code proper.
	 */
	jsr	_start_kernel			/* Start Linux kernel */

_exit:
	jmp	_exit				/* Should never get here */

/*****************************************************************************/
