/*
 * ip_nettel.c
 */

#ifdef CONFIG_NETtel

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <unistd.h>

#include "ip_nettel.h"

#define MAX_CONFIG_LINE_SIZE	128

/* local prototypes */
int writeConfig(char *filename, char *keyword, char *value);
#if 0
int filecopy(char *from, char *to);
#endif
int writeOptionsLine(char *filename, char *line);
void launch_script(char *filename);
int find_in_line(char *filename /* hay */, char *needle);

/* write the IP address to /etc/config/config so the user can tell what IP address the
 * NETtel got. Secondly, launch: ipfwadm -F -a m -S 0/0 -D 0/0 */
int ipfwadm_rules(char *ifname, u_int32_t yiaddr) {
	char buf[32];
	FILE *in;
	pid_t pid;
	char tmp[MAX_CONFIG_LINE_SIZE];
	struct in_addr inp;
	
	inp.s_addr = yiaddr;
		
	sprintf(buf, "ip%s\0", ifname);
	writeConfig("/etc/config/config", buf, (char *) inet_ntoa(inp));
#if 0
	sprintf(tmp, "ipfwadm -F -a m -S %s/24 -D 0/0\0", inet_ntoa(inp));
	writeOptionsLine("/etc/config/ipfwrules", tmp);
#else
	writeOptionsLine("/etc/config/ipfwrules", "ipfwadm -F -a m -S 0/0 -D 0/0");
#endif
	launch_script("/etc/config/ipfwrules");
			
	/* burn into flash */
	/* get the pid of flatfsd */
	in = fopen("/var/log/flatfsd.pid", "r");
	if(in != NULL) {
		if(fread(&pid, sizeof(pid_t), 1, in) > 0) {
			kill(pid, 10);
		}
		fclose(in);
	}
	return 0;
}

/* writeOptionsLine */
int writeOptionsLine(char *filename, char *line) {
        FILE *in;
        char line_buffer[128];

	if(find_in_line(filename, line) == 1)
		return 0;

        in = fopen(filename, "a");
        sprintf(line_buffer, "%s\n\0", line);
        fputs(line_buffer, in);
        fclose(in);
        return 0;
}


int writeConfig(char *filename, char *keyword, char *value) {
	FILE *in;
	FILE *out;
	
	char line_buffer[64];
	char tmp[64];

	in = fopen(filename, "r");
	out = fopen("/etc/config/tmp", "w");
	
	while((fgets(line_buffer, 63, in)) != NULL) {
		if(sscanf(line_buffer, "%s", tmp) > 0) {
			if(!strcmp(tmp, keyword)) {
				/* don't write this line to the tmp file as we
				* are going to replace it later */
			} else {
				fputs(line_buffer, out);
			}
		}
	}	
	fclose(in);
	
	sprintf(tmp, "%s %s\n\0", keyword, value);
	fputs(tmp, out);
	fclose(out);

#if 0
	filecopy("/etc/config/tmp", filename);
#else
	unlink(filename);
	link("/etc/config/tmp", filename);
#endif
	unlink("/etc/config/tmp");
	
	return 0;
}

#if 0
int filecopy(char *from, char *to) {
	unsigned char	buf[1024];
	int		n, rc;
	int		fdfrom, fdto;

	if ((fdfrom = open(from, O_RDONLY)) < 0)
		return(-1);
	if ((fdto = open(to, O_WRONLY | O_TRUNC | O_CREAT, 0644)) < 0) {
		close(fdfrom);
		return(-1);
	}

	for (rc = 0;;) {
		if ((n = read(fdfrom, &buf[0], sizeof(buf))) < 0) {
			rc = -1;
			break;
		}
		if(n == 0)
			break;
		if (write(fdto, (void *) &buf[0], n) != n)
			break;
	}
	close(fdto);
	close(fdfrom);
	return(rc);
}
#endif

int addGateway(u_int32_t gateway) {
	/* the correct way to do this would be via an ioctl
	 * call.. but it doesn't appear to work hence this
	 * nasty hack :-) */
	char buf[16];
	char *route_argv[6];
	int s, route_argc = 0;
	pid_t pid;
	
	if((pid = vfork()) == 0) { /* child */
		route_argv[route_argc++] = "/bin/route";
		route_argv[route_argc++] = "add";
		route_argv[route_argc++] = "default";
		route_argv[route_argc++] = "gw";
		sprintf(buf, "%s\0", inet_ntoa(gateway));
		route_argv[route_argc++] = buf;
		route_argv[route_argc] = NULL;
		execvp("/bin/route", route_argv);	
		exit(0);
	} else if(pid > 0) {
		waitpid(pid, &s, 0);
	}
	
	return 0;
}


void launch_script(char *filename) {
        char *argv[3];
        int s, argc = 0;
        pid_t pid;

        if((pid = vfork()) == 0) {
                argv[argc++] = "/bin/sh";
                argv[argc++] = filename;
                argv[argc] = NULL;
                execvp("/bin/sh", argv);
                exit(0);
        } else if (pid > 0) {
                waitpid(pid, &s, 0);
        }
}


int find_in_line(char *filename /* hay */, char *needle) {
        FILE *in;
        char line_buffer[MAX_CONFIG_LINE_SIZE]; /* 128 */

        in = fopen(filename, "r");
        if(in == NULL)
                return -1;

        while((fgets(line_buffer, MAX_CONFIG_LINE_SIZE - 1, in)) != NULL) {
                if(strstr(line_buffer, needle) != NULL) {
                        fclose(in);
                        return 1;
                }
        }
        return 0;
}


#endif /* CONFIG_NETtel */
