/* simpleinit.c - poe@daimi.aau.dk */
/* Version 1.21 */

/* gerg@moreton.com.au -- modified for direct console support DEC/1999 */

#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/termios.h>
#include <utmp.h>
#ifdef SHADOW_PWD
#include <shadow.h>
#endif

#include "pathnames.h"

#define CMDSIZ 80	/* max size of a line in inittab */
#define NUMCMD 16	/* max number of lines in inittab */
#define NUMTOK 10	/* max number of tokens in inittab command */

#define getdtablesize() (64)

#define RUN_RC
#define TZFILE "/etc/TZ"
char tzone[CMDSIZ];
/* #define DEBUGGING */

/* Define this if you want init to ignore the termcap field in inittab for
   console ttys. */
/* #define SPECIAL_CONSOLE_TERM */

#define ever (;;)

struct initline {
	pid_t	pid;
	char	tty[10];
	char	termcap[30];
	char	*toks[NUMTOK];
	char	line[CMDSIZ];
};

struct initline inittab[NUMCMD];
int numcmd;
int stopped = 0;	/* are we stopped */

int do_rc(), do_rcfile();
void spawn(), hup_handler(), read_inittab(), read_initfile();
void tstp_handler(), int_handler(), cont_handler();
void set_tz(), write_wtmp();
void make_ascii_tty(), make_console();
int boot_single();

/* Keep track of console device, if any... */
char	*console_device = NULL;
int	console_baud = CONSOLE_BAUD_RATE;


void err(char *s)
{
	int fd;
	
	if (console_device == NULL) return;
	if((fd = open(console_device, O_WRONLY)) < 0) return;

	write(fd, "init: ", 6);	
	write(fd, s, strlen(s));
	close(fd);
}

void
enter_single()
{
    struct termios tty;
    pid_t pid;
    int i;

    err("Booting to single user mode.\n");
    if((pid = vfork()) == 0) {
	/* the child */
	execl(_PATH_BSHELL, _PATH_BSHELL, NULL);
	err("exec of single user shell failed\n");
    } else if(pid > 0) {
	while(wait(&i) != pid) /* nothing */;
    } else if(pid < 0) {
	err("fork of single user shell failed\n");
    }
    unlink(_PATH_SINGLE);
}

int main(int argc, char *argv[])
{
	int 	vec,i;
	pid_t	pid;
	char	*sp;

#ifdef SET_TZ
	set_tz();
#endif
	signal(SIGTSTP, tstp_handler);
	signal(SIGCONT, cont_handler);
	signal(SIGINT, int_handler);

	/* 
	 * start up in single user mode if /etc/singleboot exists or if
	 * argv[1] is "single".
	 */
	if(boot_single(0, argc, argv)) enter_single();

#ifdef RUN_RC
	/* Register console if defined by boot */
	if ((console_device = getenv("CONSOLE"))) {
		unsetenv("CONSOLE");
		if ((sp = strchr(console_device, ','))) {
			*sp++ = 0;
			i = atoi(sp);
			if ((i > 0) && (i <= 460800))
				console_baud = i;
		}
	}
	make_ascii_tty();

	/*If we get a SIGTSTP before multi-user mode, do nothing*/
	while(stopped)	
		pause();
	if(do_rc() != 0 && boot_single(1, argc, argv) && !stopped)
		enter_single();
	while(stopped)	/*Also if /etc/rc fails & we get SIGTSTP*/
		pause();
#endif

	write_wtmp();	/* write boottime record */

	for(i = 0; i < NUMCMD; i++)
		inittab[i].pid = -1;

	read_inittab();

#ifdef DEBUGGING
	for(i = 0; i < numcmd; i++) {
		char **p;
		p = inittab[i].toks;
		printf("toks= %s %s %s %s\n",p[0], p[1], p[2], p[3]);
		printf("tty= %s\n", inittab[i].tty);
		printf("termcap= %s\n", inittab[i].termcap);
	}
	/*exit(0);*/
#endif
	signal(SIGHUP, hup_handler);

	for(i = 0; i < getdtablesize(); i++) close(i);

	for(i = 0; i < numcmd; i++)
		spawn(i);
	
	for ever {
		pid = wait(&vec);

		/* clear utmp entry, and append to wtmp if possible */
		{
		    struct utmp *ut;
		    int ut_fd;

		    utmpname(_PATH_UTMP);
		    setutent();
		    while((ut = getutent())) {
			if(ut->ut_pid == pid) {
			    time(&ut->ut_time);
			    memset(&ut->ut_user, 0, UT_NAMESIZE);
			    memset(&ut->ut_host, 0, sizeof(ut->ut_host));
			    ut->ut_type = DEAD_PROCESS;
			    ut->ut_pid = 0;
			    ut->ut_addr = 0;
			    endutent();
			    pututline(ut);
			    if((ut_fd = open(_PATH_WTMP, O_APPEND|O_WRONLY)) >= 0) {
				flock(ut_fd, LOCK_EX|LOCK_NB);
				write(ut_fd, ut, sizeof(struct utmp));
				flock(ut_fd, LOCK_UN|LOCK_NB);
				close(ut_fd);
			    }
			    break;
			}
		    }
		    endutent();
		}

		for(i = 0; i < numcmd; i++) {
			if(pid == inittab[i].pid || inittab[i].pid < 0) {
				if(stopped) inittab[i].pid = -1;
				else spawn(i);
				/*break;*/
			}
		}
	}
}	

#define MAXTRIES 3 /* number of tries allowed when giving the password */

/*
 * return true if we should boot up in singleuser mode. If argv[i] is 
 * "single" or the file /etc/singleboot exists, then singleuser mode should
 * be entered. If /etc/securesingle exists ask for root password first.
 */
int boot_single(int singlearg, int argc, char *argv[])
{
	char *pass, *rootpass = NULL;
	struct passwd *pwd;
	int i;

	for(i = 1; i < argc; i++) {
	    if(argv[i] && !strcmp(argv[i], "single")) singlearg = 1;
	}

	if(access(_PATH_SINGLE, 04) == 0 || singlearg) {
		if(access(_PATH_SECURE, 04) == 0) {
			if((pwd = getpwnam("root")) || (pwd = getpwuid(0)))
			  rootpass = pwd->pw_passwd;
			else
			  return 1; /* a bad /etc/passwd should not lock out */

			for(i = 0; i < MAXTRIES; i++) {
				pass = getpass("Password: ");
				if(pass == NULL) continue;
				
				if(!strcmp(crypt(pass, rootpass), rootpass)) {
					return 1;
				}

				puts("\nWrong password.\n");
			}
		} else return 1;
	}
	return 0;
}

/*
 * run /etc/rc. The environment is passed to the script, so the RC environment
 * variable can be used to decide what to do. RC may be set from LILO.
 */
int do_rc()
{
	do_rcfile(_PATH_RC);
#if defined(CONFIG_NETtel) || defined(CONFIG_eLIA) || defined(CONFIG_DISKtel)
	do_rcfile(_PATH_CONFIGRC);
	do_rcfile(_PATH_IPFWRULES);
#endif
}

int do_rcfile(char *rcfile)
{
	pid_t pid;
	int stat, st;
	
	if((pid = vfork()) == 0) {
		/* the child */
		char *argv[3];

		close(0);
		argv[0] = _PATH_BSHELL;
#if 1
		argv[1] = rcfile;
		argv[2] = (char *)0;
		execv(_PATH_BSHELL, argv);
		err("exec rc failed\n");
		return(2);
#else
		argv[1] = (char *)0;
		if(open(rcfile, O_RDONLY, 0) == 0) {
			execv(_PATH_BSHELL, argv);
			err("exec rc failed\n");
			return(2);
		}
		err("open of rc file failed\n");
		return(1);
#endif
	} else if(pid > 0) {
		/* parent, wait till rc process dies before spawning */
		while(wait(&stat) != pid) /* nothing */;
	} else if(pid < 0) {
		err("fork of rc shell failed\n");
	}
	st = WEXITSTATUS(stat);
	return st;
}

void spawn(int i)
{
	pid_t pid;
	int j;
	
	if((pid = vfork()) < 0) {
		inittab[i].pid = -1;
		err("fork failed\n");
		return;
	}
	if(pid) {
		/* this is the parent */
		inittab[i].pid = pid;
		return;
	} else {
		/* this is the child */
		char term[40];
#ifdef SET_TZ
		char tz[CMDSIZ];
#endif
		char *env[3];
		
		setsid();
		for(j = 0; j < getdtablesize(); j++)
			(void) close(j);

		if (strcmp(inittab[i].tty, "console") == 0)
			make_console();

		(void) sprintf(term, "TERM=%s", inittab[i].termcap);
		env[0] = term;
		env[1] = (char *)0;
#ifdef SET_TZ
		(void) sprintf(tz, "TZ=%s", tzone);
		env[1] = tz;
#endif
		env[2] = (char *)0;

		execve(inittab[i].toks[0], inittab[i].toks, env);
		err("exec failed\n");
		sleep(5);
		_exit(1);
	}
}

void read_inittab()
{
	numcmd = 0;

	/* Fake an inittab entry if boot console defined */
	if (console_device && strcmp(console_device, "/dev/null")) {
		strcpy(inittab[numcmd].tty, "console");
		strcpy(inittab[numcmd].termcap, "linux");
		inittab[numcmd].toks[0] = "/bin/sh";
		inittab[numcmd].toks[1] = NULL;
		numcmd++;
	}

	read_initfile(_PATH_INITTAB);

#if defined(CONFIG_NETtel) || defined(CONFIG_eLIA) || defined(CONFIG_DISKtel)
	read_initfile(_PATH_CONFIGTAB);
#endif

	if (numcmd == 0)
		_exit(1);
}

void read_initfile(char *initfile)
{
	FILE *f;
	char buf[CMDSIZ];
	int i,j,k;
	char *ptr, *getty;
#ifdef SPECIAL_CONSOLE_TERM
	char tty[50];
	struct stat stb;
#endif
	char *termenv, *getenv();
	
	termenv = getenv("TERM");	/* set by kernel */
	/* termenv = "vt100"; */
			
	i = numcmd;

	if(!(f = fopen(initfile, "r"))) {
		err("cannot open inittab\n");
		return;
	}

	while(!feof(f) && i < NUMCMD - 2) {
		if(fgets(buf, CMDSIZ - 1, f) == 0) break;
		buf[CMDSIZ-1] = 0;

		for(k = 0; k < CMDSIZ && buf[k]; k++) {
			if(buf[k] == '#') { 
				buf[k] = 0; break; 
			}
		}

		if(buf[0] == 0 || buf[0] == '\n') continue;

		(void) strcpy(inittab[i].line, buf);
		(void) strtok(inittab[i].line, ":");
		(void) strncpy(inittab[i].tty, inittab[i].line, 10);
		inittab[i].tty[9] = 0;
		(void) strncpy(inittab[i].termcap,
				strtok((char *)0, ":"), 30);
		inittab[i].termcap[29] = 0;

		getty = strtok((char *)0, ":");
		(void) strtok(getty, " \t\n");
		inittab[i].toks[0] = getty;
		j = 1;
		while((ptr = strtok((char *)0, " \t\n")))
			inittab[i].toks[j++] = ptr;
		inittab[i].toks[j] = (char *)0;

#ifdef SPECIAL_CONSOLE_TERM
		/* special-case termcap for the console ttys */
		(void) sprintf(tty, "/dev/%s", inittab[i].tty);
		if(!termenv || stat(tty, &stb) < 0) {
			err("no TERM or cannot stat tty\n");
		} else {
			/* is it a console tty? */
			if(major(stb.st_rdev) == 4 && minor(stb.st_rdev) < 64) {
				strncpy(inittab[i].termcap, termenv, 30);
				inittab[i].termcap[29] = 0;
			}
		}
#endif

		i++;
	}
	fclose(f);

	numcmd = i;
}

void hup_handler()
{
	int i,j;
	int oldnum;
	struct initline	savetab[NUMCMD];
	int had_already;

	(void) signal(SIGHUP, SIG_IGN);

	memcpy(savetab, inittab, NUMCMD * sizeof(struct initline));
	oldnum = numcmd;		
	read_inittab();
	
	for(i = 0; i < numcmd; i++) {
		had_already = 0;
		for(j = 0; j < oldnum; j++) {
			if(!strcmp(savetab[j].tty, inittab[i].tty)) {
				had_already = 1;
				if((inittab[i].pid = savetab[j].pid) < 0)
					spawn(i);
			}
		}
		if(!had_already) spawn(i);
	}
	
	/* Kill off processes no longer needed */
	for(i = 0; i < oldnum; i++) {
		had_already = 0;
		for(j = 0; j < numcmd; j++) {
			if(!strcmp(inittab[j].tty, savetab[i].tty))
				break;
		}
		if (j >= numcmd)
			kill(savetab[i].pid, SIGTERM);
	}

	(void) signal(SIGHUP, hup_handler);
}

void tstp_handler()
{
	stopped++;
	signal(SIGTSTP, tstp_handler);
}

void cont_handler()
{
	stopped = 0;
	signal(SIGCONT, cont_handler);
}

void int_handler()
{
	/*
	 * After Linux 0.96b PL1, we get a SIGINT when
	 * the user presses Ctrl-Alt-Del...
	 */

	int pid;
	
	sync();
	sync();
	if((pid = vfork()) == 0) {
		/* reboot properly... */
		execl(_PATH_REBOOT, _PATH_REBOOT, (char *)0);
		reboot(0xfee1dead, 672274793, 0x1234567);
	} else if(pid < 0)
		/* fork failed, try the hard way... */
		reboot(0xfee1dead, 672274793, 0x1234567);
}

void set_tz()
{
#if 0
	FILE *f;
	int len;

	if((f=fopen(TZFILE, "r")) == (FILE *)NULL) return;
	fgets(tzone, CMDSIZ-2, f);
	fclose(f);
	if((len=strlen(tzone)) < 2) return;
	tzone[len-1] = 0; /* get rid of the '\n' */
	setenv("TZ", tzone, 0);
#endif
}

void write_wtmp()
{
#if 0
    int fd;
    struct utmp ut;
    
    memset((char *)&ut, 0, sizeof(ut));
    strcpy(ut.ut_line, "~");
    memset(ut.ut_name, 0, sizeof(ut.ut_name));
    time(&ut.ut_time);
    ut.ut_type = BOOT_TIME;
    
    if((fd = open(_PATH_WTMP, O_WRONLY|O_APPEND)) >= 0) {
	flock(fd, LOCK_EX|LOCK_NB); /* make sure init won't hang */
	write(fd, (char *)&ut, sizeof(ut));
	flock(fd, LOCK_UN|LOCK_NB);
	close(fd);
    }
#endif
}     

void make_ascii_tty()
{
	struct termios tty;

	tcgetattr(0, &tty);
        tty.c_iflag &= ~(INLCR|IGNCR|IUCLC);
        tty.c_iflag |= ICRNL;
        tty.c_oflag &= ~(OCRNL|OLCUC|ONOCR|ONLRET|OFILL);
        tty.c_oflag |= OPOST|ONLCR;
        tty.c_cflag |= CLOCAL;
        tty.c_lflag  = ISIG|ICANON|ECHO|ECHOE|ECHOK|ECHOCTL|ECHOKE;

	cfsetospeedn(&tty, console_baud);

        tty.c_cc[VINTR]  = 3;
        tty.c_cc[VQUIT]  = 28;
        tty.c_cc[VERASE] = 8/*127*/;
        tty.c_cc[VKILL]  = 24;
        tty.c_cc[VEOF]   = 4;
        tty.c_cc[VTIME]  = 0;
        tty.c_cc[VMIN]   = 1;
        tty.c_cc[VSTART] = 17;
        tty.c_cc[VSTOP]  = 19;
        tty.c_cc[VSUSP]  = 26;

	tcsetattr(0, TCSANOW, &tty);
}

void make_console()
{
	int j;
	struct termios tty;

	if (!console_device)
		return;
	if (open(console_device, O_RDWR|O_NONBLOCK) == -1)
		return;
	fcntl(0, F_SETFL, 0);
	dup(0); 
	dup(0);
	make_ascii_tty();
	j = ioctl(0, TIOCSCTTY, (char*)0);
}

