#include "tinylogin.def.h"
#include <stdio.h>
#include <grp.h>
#include <string.h>
#include <pwd.h>
#include "internal.h"
#include "gshadow_.h"

#include "group.h"

#ifdef TLG_FULL_GROUP_SUPPORT

#define GROUP_DBG 0

#if GROUP_DBG
#define dbgprint(x) printf x
#else
#define dbgprint(x)
#endif /*GROUP_DBG*/


char *
GetGrpLine(FILE *fd, char *buf, int oldSize, int *newSize);

char *
fgets_m(
	char *str,
	int n,
	int *actual,
	FILE *fd
	);

char **
CopyData(char **src, int n);




/*
 *Allow the full on modification of groups.
 */

int
writegrent(const struct group *gr)
/*
desc:	This can be used to write group information to the group file.
		Another function must be used to write out shadow group information.
		It does no checks at all.
*/
{
	FILE *in;
	FILE *out;
	char *line;
	int chs = 80;
	int n;
	int fnd = 0;
	char *grp;
	int written = 0;
	int size = 0;

	line = (char *)malloc(chs*sizeof(char));

	if(!line){
		return -1;
	}
	in = fopen(GRP_FILE, "r");
	out = fopen("tmpfile", "w");
	while((line = GetGrpLine(in, line, size, &size)) != NULL){
/*		dbgprint(("WRITEGRENT: - %s\n", line));*/

		grp = strtok(line, ":");
		if(strcmp(gr->gr_name, grp) == 0){
			written = 1;
			/*Write all of the group information*/
			writegr(out, gr);
		}
		else
		{
			line[strlen(line)] = ':';
			fputs(line, out);
			fputs("\n", out);
		}
	}

	/*write out the information now*/
	if(!written)
		writegr(out, gr);
	/*
	copy the file back;
	*/
	fclose(out);
#if 0
	unlink(GRP_FILE);
	link("/etc/config/tmpfile", GRP_FILE);
	unlink("/etc/config/tmpfile");
#endif
	return 0;
}



char *
GetGrpLine(FILE *fd, char *buf, int oldSize, int *newSize)
{
	int size = 0;
	char * oldBuf;

	if(fgets_m(buf, oldSize, &size, fd) == NULL)
	{
		return NULL;
	}

	if(oldSize <= size)
	{
		//
		//unread the data we just read (these values appear to work!!)
		//
		fseek(fd, ftell(fd) - size, SEEK_SET);

		//
		//realloc the data.
		//
		oldBuf = buf;

		buf = (char *)realloc(buf, size + sizeof(char));

		if(buf == NULL)
		{
			return NULL;
		}

		oldSize = size;

		//
		//re-read the data.
		//
		if(fgets_m(buf, oldSize, &size, fd) == NULL)
		{
			//
			//obviously something else went badly wrong.
			//
			return NULL;
		}
	}
	return buf;
}


char *
fgets_m(
	char *str,
	int n,
	int *actual,
	FILE *fd
	)
{
	int i = 0;  //num chars currently read.
	char ch;

	while(((ch = fgetc(fd)) != EOF) && (ch != '\n'))
	{
		if(ch == '\n')
		{
			break;
		}

		if(i < n)
		{
			str[i] = ch;
		}
		i++;
	}

	*actual = i + 1;

	(i<n)?(str[i] = '\0'):(str[n-1] = '\0');

	if(i == 0)
		return NULL;
	else
		return str;
}


writegr(FILE *fd, struct group *gr)
/*
desc:	This function is used to write group information out to a file
		it assumes the file pointer is pointing somewhere useful.
*/
{
	char *tmp;
	int i = 0;

//	line[strlen[line]] = ':';

	fprintf(fd, "%s:%s:%u",gr->gr_name, gr->gr_passwd, gr->gr_gid);
	fprintf(fd,":");
	if((gr->gr_mem != NULL) && (gr->gr_mem[0] != '\0'))		/*write out all members*/
	{
		fprintf(fd, "%s", gr->gr_mem[0]);
		for(i=1;gr->gr_mem[i] != '\0';i++)
		{
			fprintf(fd, ",%s", gr->gr_mem[i]);
		}
	}

	fprintf(fd, "\n");
}

int gr_addusr(struct group *grp, char *usr)
/*
desc:	simply adds the given user to the group structure.
		if the user is already in the group then nothing is done.
*/
{
	int freestuff = 0;
	int i;
	int size = 0;
	char ** res;

#if 0
	if(grp->gr_mem != ((char *)grp + sizeof(struct group)))
	{
		dbgprint(("gr_addusr:  freeeing old memory\n"));
		sleep(1);
		freestuff = 1;
	}
#endif
	for(i=0;grp->gr_mem[i] != NULL;i++){/* simply get the number of groups*/
		dbgprint(("Mem%i - %s\n",i,grp->gr_mem[i]));
		size++;
	}
	for(i=0;grp->gr_mem[i] != NULL;i++){/*is the group already present*/
		if(strcmp(grp->gr_mem[i], usr) == 0)
			return 1;		/*already present*/
	}
	res = CopyData(grp->gr_mem, 2);	/*copy the grp into into a new structure, and allocate extra space*/
	/*copy in the new information*/
	grp->gr_mem = res;
	res[size] = (char *)malloc(strlen(usr));
	strcpy(res[size], usr);
	res[size+1] = NULL;
	grp->gr_mem = res;
	return 0;

}

char **
CopyData(char **src, int n)
/*
desc:	copies a string ** and allocs n extra data entries
*/
{
	int i=0;
	char **res;

	for(i=0;src[i] != 0;i++) {printf("%s\n",src[i]);}
	res = (char **)malloc(sizeof(char *) * i + n + 1);
	dbgprint(("\n\n"));
	if(res == NULL)
		return res;
	for(i=0;src[i] != 0;i++){
		res[i] = (char *)malloc(strlen(src[i]) + 1);
		dbgprint(("%s\n",src[i]));
		strcpy(res[i], src[i]);
	}

	dbgprint(("CopyData: Exiting\n"));

	return res;
}

#ifdef TLG_FEATURE_SHADOWPASSWDS
sgr_addusr(struct sgrp *sgrp, char *name)
{
	int freestuff = 0;
	int i;
	int size = 0;
	char ** res;

	/*
	MN - BUG BUG, should we even bother unmallocing?!!
	*/
/*	if(grp->gr_mem != ((char *)grp + sizeof(struct group)))
	{
		dbgprint(("gr_addusr:  freeeing old memory\n"));
		sleep(1);
		freestuff = 1;
	}*/

	dbgprint(("sgr_addusr - Adding a user to a sh group\n"));
	for(i=0;sgrp->sg_mem[i] != NULL;i++)
	{
		size++;
	}

	/*
	check to make sure that the user is not already in the group.
	*/
	for(i=0;sgrp->sg_mem[i] != NULL;i++){
		if(strcmp(sgrp->sg_mem[i], name) == 0)
			return 1;		/*already present*/
	}

	/*
	Need 1 for extra element and 1 for null terminator.
	*/
	res = CopyData(sgrp->sg_mem, 2);

	/*if(freestuff)
	{
		dbgprint(("Freeing data\n"));
//		free(grp->gr_mem);
	}
*/
	sgrp->sg_mem = res;
	/*
	grp->gr_mem[size] = (char *)malloc(strlen(usr));
	*/
	res[size] = (char *)malloc(strlen(name));
	/*
	strcpy(grp->gr_mem[size], usr);
	*/
	strcpy(res[size], name);
	/*
	grp->gr_mem[size + 1]=0;
	*/
	res[size+1] = NULL;

	sgrp->sg_mem = res;
	return 0;
}

int
writesgrent(struct sgrp *grp)
{
	/*write out the group structure*/
	FILE *in;
	FILE *out;
	char *line;
	int chs = 80;
	int n;
	int fnd = 0;
	char *gr;
	int written = 0;
	int size = 0;

	line = (char *)malloc(chs*sizeof(char));
	if(!line)
		return 0;//errno;

	
	in = fopen(SGROUP_FILE, "r");
	if(in == NULL)
	{
		return 0;//errno;
	}
	out = fopen("tmpfile2","w");
	if(out == NULL)
	{
		return 0;//errno;
	}
	while((line = GetGrpLine(in, line, size, &size)) != NULL){

		gr = strtok(line, ":");
		if(strcmp(grp->sg_name, gr) == 0){
			written = 1;
			/*Write all of the group information*/
//			writegr(out, gr);
			putsgent(grp, out);	
		}
		else
		{
			line[strlen(line)] = ':';
			fputs(line, out);
			fputs("\n", out);
		}
	}

	fclose(out);
#if 0
	unlink(SGROUP_FILE);
	link("/etc/config/tmpfile2", SGROUP_FILE);
	unlink("/etc/config/tmpfile2");
#endif

}

#endif /*TLG_FEATURE_SHADOWPASSWDS*/

int AddUserToGroup(char *userName, char *grpName)
/*
	Simply add the user to the above group.
	0 for failure,
	1 for success-a-roony.
*/
{
	struct passwd *pw;
	struct group *gr;
	int i;

	pw = getpwnam(userName);
	if(pw == NULL){
		return 0;
	}

	gr = getgrnam(grpName);
	if(gr == NULL){
		return 0;
	}

	/*
	Is the user already a member of this group
	*/
#if 1  /*This might save us searching all of the group members*/
	if(pw->pw_gid == gr->gr_gid){
		return 1;
	}
#endif
	for(i=0;gr->gr_mem[i] != NULL;i++){
		if(strcmp(gr->gr_mem[i], pw->pw_name) != 0)
			return 1;
	}
	/*write out the group stuff*/

	
	return 1;
}

#endif //TLG_FULL_GROUP_SUPPORT