/*
 * Copyright 1990 - 1994, Julianne Frances Haugh
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "internal.h"

//#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>


/* Newer versions of Linux libc already have shadow support.  */
/*#if defined(SHADOWGRP) && !defined(HAVE_SHADOWGRP)	/*{*/
/*
#include "rcsid.h"
RCSID("$Id: gshadow.c,v 1.1 2000/06/01 00:43:08 matthewn Exp $")

#include <stdio.h>
#include "prototypes.h"
#include "defines.h"

*/
#ifdef TLG_FEATURE_SHADOWPASSWDS

#define	MAXMEM	1024

static	FILE	*shadow;
static	char	sgrbuf[BUFSIZ*4];
static	char	*members[MAXMEM+1];
static	char	*admins[MAXMEM+1];
static	struct	sgrp	sgroup;

extern	char	*fgetsx();
extern	int	fputsx();

#define	FIELDS	4


static char **
list(char *s, char **l)
{
	int	nmembers = 0;

	while (s && *s) {
		l[nmembers++] = s;
		if ((s = strchr (s, ',')))
			*s++ = '\0';
	}
	l[nmembers] = (char *) 0;
	return l;
}

void
setsgent(void)
{
	if (shadow)
		rewind (shadow);
	else
		shadow = fopen(SGROUP_FILE, "r");

	/*
	 * Attempt to open the DBM files if they have never been opened
	 * and an error has never been returned.
	 */
}

void
endsgent(void)
{
	if (shadow)
		(void) fclose (shadow);

	shadow = (FILE *) 0;
}

struct sgrp *
sgetsgent(const char *string)
{
	char	*fields[FIELDS];
	char	*cp;
	int	i;

	strncpy (sgrbuf, string, (int) sizeof sgrbuf - 1);
	sgrbuf[sizeof sgrbuf - 1] = '\0';

	if ((cp = strrchr (sgrbuf, '\n')))
		*cp = '\0';

	/*
	 * There should be exactly 4 colon separated fields.  Find
	 * all 4 of them and save the starting addresses in fields[].
	 */

	for (cp = sgrbuf, i = 0;i < FIELDS && cp;i++) {
		fields[i] = cp;
		if ((cp = strchr (cp, ':')))
			*cp++ = '\0';
	}

	/*
	 * If there was an extra field somehow, or perhaps not enough,
	 * the line is invalid.
	 */

	if (cp || i != FIELDS)
		return 0;

	sgroup.sg_name = fields[0];
	sgroup.sg_passwd = fields[1];
	sgroup.sg_adm = list (fields[2], admins);
	sgroup.sg_mem = list (fields[3], members);

	return &sgroup;
}

/*
 * fgetsgent - convert next line in stream to (struct sgrp)
 *
 * fgetsgent() reads the next line from the provided stream and
 * converts it to a (struct sgrp).  NULL is returned on EOF.
 */

struct sgrp *
fgetsgent(FILE *fp)
{
	char	buf[sizeof sgrbuf];
	char	*cp;

	if (! fp)
		return (0);

	if (fgetsx (buf, sizeof buf, fp) != (char *) 0)
	{
		if ((cp = strchr (buf, '\n')))
			*cp = '\0';
		return (sgetsgent (buf));
	}
	return 0;
}

/*
 * getsgent - get a single shadow group entry
 */

struct sgrp *
getsgent(void)
{
	if (! shadow)
		setsgent ();

	return (fgetsgent (shadow));
}

/*
 * getsgnam - get a shadow group entry by name
 */

struct sgrp *
getsgnam(const char *name)
{
	struct	sgrp	*sgrp;

	setsgent ();

	while ((sgrp = getsgent ()) != (struct sgrp *) 0) {
		if (strcmp (name, sgrp->sg_name) == 0)
			break;
	}
	if (sgrp)
		return sgrp;
	return (0);
}

/*
 * putsgent - output shadow group entry in text form
 *
 * putsgent() converts the contents of a (struct sgrp) to text and
 * writes the result to the given stream.  This is the logical
 * opposite of fgetsgent.
 */

int
putsgent(const struct sgrp *sgrp, FILE *fp)
{
	char *buf, *cp;
	int i;
	size_t size;

	if (! fp || ! sgrp)
		return -1;

	/* calculate the required buffer size */
	size = strlen(sgrp->sg_name) + strlen(sgrp->sg_passwd) + 10;
	for (i = 0; sgrp->sg_adm && sgrp->sg_adm[i]; i++)
		size += strlen(sgrp->sg_adm[i]) + 1;
	for (i = 0; sgrp->sg_mem && sgrp->sg_mem[i]; i++)
		size += strlen(sgrp->sg_mem[i]) + 1;

	buf = malloc(size);
	if (!buf)
		return -1;
	cp = buf;

	/*
	 * Copy the group name and passwd.
	 */

	strcpy (cp, sgrp->sg_name);
	cp += strlen (cp);
	*cp++ = ':';

	strcpy (cp, sgrp->sg_passwd);
	cp += strlen (cp);
	*cp++ = ':';

	/*
	 * Copy the administrators, separating each from the other
	 * with a ",".
	 */

	for (i = 0;sgrp->sg_adm[i];i++) {
		if (i > 0)
			*cp++ = ',';

		strcpy (cp, sgrp->sg_adm[i]);
		cp += strlen (cp);
	}
	*cp++ = ':';

	/*
	 * Now do likewise with the group members.
	 */

	for (i = 0;sgrp->sg_mem[i];i++) {
		if (i > 0)
			*cp++ = ',';

		strcpy (cp, sgrp->sg_mem[i]);
		cp += strlen (cp);
	}
	*cp++ = '\n';
	*cp = '\0';

	/*
	 * Output using the function which understands the line
	 * continuation conventions.
	 */

	if (fputsx(buf, fp) == EOF) {
		free(buf);
		return -1;
	}

	free(buf);
	return 0;
}

#else
extern int errno;  /* warning: ANSI C forbids an empty source file */
#endif	/*TLG_FEATURE_SHADOWPASSWDS */


char *
fgetsx(char *buf, int cnt, FILE *f)
{
	char *cp = buf;
	char *ep;

	while (cnt > 0) {
		if (fgets (cp, cnt, f) == 0) {
			if (cp == buf)
				return 0;
			else
				break;
		}
		if ((ep = strrchr (cp, '\\')) && *(ep + 1) == '\n') {
			if ((cnt -= ep - cp) > 0)
				*(cp = ep) = '\0';
		} else
			break;
	}
	return buf;
}

int
fputsx(const char *s, FILE *stream)
{
	int i;

	for (i = 0;*s;i++, s++) {
		if (putc (*s, stream) == EOF)
			return EOF;

#if 0  /* The standard getgr*() can't handle that.  --marekm */
		if (i > (BUFSIZ/2)) {
			if (putc ('\\', stream) == EOF ||
			    putc ('\n', stream) == EOF)
				return EOF;

			i = 0;
		}
#endif
	}
	return 0;
}
