#include "internal.h"

#include <fcntl.h>
#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <utime.h>
#include "../../linux/include/linux/autoconf.h"

#ifdef USE_SYSLOG
#include <syslog.h>
#endif

#include <time.h>
#include <sys/resource.h>
#include <getopt.h>

#define PASSWD_FILE "/etc/config/passwd"
#define SHADOW_FILE "/etc/config/shadow"

#ifndef CONFIG_REDUCED_MEMORY
static const char passwd_usage[] =
	"passwd [OPTION] [name]\n\n"
	"Change a user password. If no name is specified, changes the password\n"
	"for the current user.\n\n"
	"Options:\n"
	"\t-d\tDelete the password for the specified user account.\n"

	"\t-l\tLocks (disables) the specified user account.\n"
	"\t-u\tUnlocks (re-enables) the specified user account..\n"
	"\t-p\tSpecify the password on the command line (Dangerous!!).\n";
#else
static const char passwd_usage [] = 
	"passwd [OPTION] [name]\n\n";
#endif //CONFIG_REDUCED_MEMORY
//MN 
//useful debuggin stuff.
//
#define DBG_PASSWD 0

#if DBG_PASSWD
#define dbgprintf(x) printf x
#else
#define dbgprintf(x)
#endif //DBG_PWSSWD


static char crypt_passwd[128];


static int check_link_count(const char *file);
static int create_backup(const char *backup, FILE * fp);
static int new_password(const struct passwd *pw, int amroot);
static void set_filesize_limit();


extern int passwd_main(int argc, char **argv)
{
	int amroot;
	char *cp;
	char *np;
	char *name;
	char *myname;
	int flag;
	int lflg = 0;				/* -l - lock account */
	int uflg = 0;				/* -u - unlock account */
	int dflg = 0;				/* -d - delete password */
	int pflg = 0;				/* -p - specify pssword on command line*/
	char * passwd;				/* the new password */
	const struct passwd *pw;
//	char *cp;

#ifdef TLG_FEATURE_SHADOWPASSWDS
	const struct spwd *sp;
#endif							/* TLG_FEATURE_SHADOWPASSWDS */
	amroot = (getuid() == 0);
#ifdef USE_SYSLOG
	openlog("passwd", LOG_PID | LOG_CONS | LOG_NOWAIT, LOG_AUTH);
#endif
#define FLAGS "dlup:"
	while ((flag = getopt(argc, argv, "dlup:")) != EOF) {
		switch (flag) {
		case 'd':
			dflg++;
			break;
		case 'l':
			lflg++;
			break;
		case 'u':
			uflg++;
			break;
		case 'p':
			pflg++;
			passwd = optarg;
			break;
		default:
			usage(passwd_usage);
		}
	}

	//MN
	//If we are root then we don't want to get my pwent.  
	//We want to get some other dudes pwent.
	//

	if(optind != 0)
	{
		//MN TESTING
		//It appears as though we have been provided with a user name.
		//

		dbgprintf(("PASSWD: We are being called on someone elses user name\n"));

		if(amroot)
		{
			dbgprintf(("PASSWD: Trying to obtain pwd struct, %s\n", argv[optind]));
			pw = getpwnam(argv[optind]);
			if(!pw)
			{
				fprintf(stdout, "passwd: Unable to find username - %s.\n", argv[optind]);
				exit(1);
			}
			dbgprintf(("PASSWD: obtained user name: %s\n", pw->pw_name));
		}
		else
		{
			//MN NOT TESTED!!
			//If my name is the same as the passed in name then we should actually succeed.
			//
			dbgprintf(("PASSWD: Trying to obtain pwd struct, %s\n", argv[optind]));
			pw = getpwnam(argv[optind]);
			if(!pw)
			{
				fprintf(stdout, "passwd: Unable to find username - %s.\n", argv[optind]);
				exit(1);
			}
			if (pw->pw_uid != getuid())
			{
				fprintf(stdout, "passwd: You are not root and can only change YOUR password.\n");
				usage(passwd_usage);
				exit(1);
			}
		}
	}
	else
	{
		pw = (struct passwd *) get_my_pwent();
		if (!pw) {
			fprintf(stdout, "passwd: Cannot determine your user name.\n");
			exit(1);
		}
	}

	name = (char *) xstrdup(pw->pw_name);

	//MN - TO DO!!
	//I don't like the below, I might put it back when I think about it some more.
	//It certainly won't work properly with my code above.
	//

/*	if (optind < argc) {

		name = argv[optind];
	} else {
		name = myname;
	}
	if ((lflg || uflg || dflg) && (optind >= argc || !amroot)) {
		usage(passwd_usage);
	}
	pw = getpwnam(name);
	if (!pw) {
		fprintf(stdout, "passwd: Unknown user %s\n", name);
		exit(1);
	}
	if (!amroot && pw->pw_uid != getuid()) {
		fprintf(stdout, "passwd: Permission denied.\n");
#ifdef USE_SYSLOG
		syslog(LOG_WARNING, "can't change pwd for `%s'", name);
		closelog();
#endif
		exit(1);
	}*/
#ifdef TLG_FEATURE_SHADOWPASSWDS
	dbgprintf(("PASSWD: Getting the shadow password stuff\n"));
	sp = getspnam(name);
	if (!sp) {
		sp = (struct spwd *) pwd_to_spwd(pw);
	}
	cp = sp->sp_pwdp;
	np = sp->sp_namp;
#else
	cp = pw->pw_passwd;
	np = name;
#endif							/* TLG_FEATURE_SHADOWPASSWDS */

	STRFCPY(crypt_passwd, cp);
	if (!(dflg || lflg || uflg || pflg)) {
		if (!amroot) {
			if (cp[0] == '!') {
				fprintf(stdout,
						"The password for `%s' cannot be changed.\n", np);
#ifdef USE_SYSLOG
				syslog(LOG_WARNING, "password locked for `%s'", np);

				closelog();
#endif
				exit(1);
			}
		}
		printf("Changing password for %s\n", name);
		if (new_password(pw, amroot)) {
			fprintf(stdout, "The password for %s is unchanged.\n", name);
#ifdef USE_SYSLOG
			closelog();
#endif
			exit(1);
		}
	} else if (lflg) {
		if (crypt_passwd[0] != '!') {
			memmove(&crypt_passwd[1], crypt_passwd,
					sizeof crypt_passwd - 1);
			crypt_passwd[sizeof crypt_passwd - 1] = '\0';
			crypt_passwd[0] = '!';
		}
	} else if (uflg) {
		if (crypt_passwd[0] == '!') {
			memmove(crypt_passwd, &crypt_passwd[1],
					sizeof crypt_passwd - 1);
		}
	} else if (dflg) {
		crypt_passwd[0] = '\0';
	} else if (pflg){
		cp = pw_encrypt(passwd, crypt_make_salt());
//		bzero(pass, sizeof pass);
		STRFCPY(crypt_passwd, cp);
	}
	
	set_filesize_limit(30000);
	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	umask(077);
	if (setuid(0)) {
		fprintf(stdout, "Cannot change ID to root.\n");
#ifdef USE_SYSLOG
		syslog(LOG_ERR, "can't setuid(0)");
		closelog();
#endif
		exit(1);
	}
	if (update_passwd(pw, crypt_passwd)) {
#ifdef USE_SYSLOG
		syslog(LOG_INFO, "password for `%s' changed by user `%s'", name,
			   myname);
#endif
		printf("Password changed.\n");
	} else {
#ifdef USE_SYSLOG
		syslog(LOG_WARNING,
			   "an error occurred updating the password file");
#endif
		printf("An error occurred updating the password file.\n");
	}
#ifdef USE_SYSLOG
	closelog();
#endif
	exit(0);
}



extern int update_passwd(const struct passwd *pw, char *crypt_pw)
{
	char file[1024];
	char lock[1024];
	char filename[1024];
	char buf[32];
	char buffer[80];
	char username[32];
	char *pw_rest;
	int pid;
	int len;
	int has_shadow = 0;
	int fd;
	int mask;
	int continued;
	FILE *fp;
	FILE *out_fp;
	struct stat sb;

	if (access(SHADOW_FILE, F_OK) == 0) {
		has_shadow = 1;
	}
	if (has_shadow) {
		// kwonsk: snprintf(filename, sizeof filename, "%s", SHADOW_FILE);
		sprintf(filename, "%s", SHADOW_FILE);
	} else {
		// kwonsk:snprintf(filename, sizeof filename, "%s", PASSWD_FILE);
		sprintf(filename, "%s", PASSWD_FILE);
	}
	// kwonsk:snprintf(file, sizeof file, "%s.%ld", filename, (long) getpid());
	// kwonsk:snprintf(lock, sizeof lock, "%s.lock", filename);
	sprintf(file, "%s.%ld", filename, (long) getpid());
	sprintf(lock, "%s.lock", filename);
	if ((fd = open(file, O_CREAT | O_EXCL | O_WRONLY, 0600)) == -1) {
		return 0;
	}

	pid = getpid();
	// kwonsk:snprintf(buf, sizeof buf, "%d", pid);
	sprintf(buf, "%d", pid);
	len = strlen(buf) + 1;
	if (write(fd, buf, len) != len) {
		close(fd);
		unlink(file);
		return 0;
	}
	close(fd);
	if (!link(file, lock) == 0) {
		unlink(file);
		return 0;
	}
	if (!check_link_count(file)) {
		unlink(file);
		return 0;
	}
	unlink(file);
// We have a lock, let's make a backup and start updating.
	if (((fp = fopen(filename, "r+")) == 0) || (fstat(fileno(fp), &sb))) {
		unlink(lock);
		return 0;
	}
	// kwonsk: snprintf(buf, sizeof buf, "%s-", filename);
	sprintf(buf, "%s-", filename);
	if (create_backup(buf, fp)) {
		fclose(fp);
		unlink(lock);
		return 0;
	}
	// kwonsk:snprintf(buf, sizeof buf, "%s+", filename);
	sprintf(buf, "%s+", filename);
	mask = umask(0777);
	out_fp = fopen(buf, "w");
	umask(mask);
	if ((!out_fp) || (fchmod(fileno(out_fp), sb.st_mode & 0777))
		|| (fchown(fileno(out_fp), sb.st_uid, sb.st_gid))) {
		fclose(fp);
		fclose(out_fp);
		unlink(buf);
		return 0;
	}

	continued = 0;
	// kwonsk: snprintf(username, sizeof username, "%s:", pw->pw_name);
	sprintf(username, "%s:", pw->pw_name);
	rewind(fp);
	while (!feof(fp)) {
		fgets(buffer, sizeof buffer, fp);
		if (!continued) {		// Check to see if we're updating this line.
			if (strncmp(username, buffer, strlen(username)) == 0) {	// we have a match.
				pw_rest = strchr(buffer, ':');
				*pw_rest++ = '\0';
				pw_rest = strchr(pw_rest, ':');
				fprintf(out_fp, "%s:%s%s", buffer, crypt_pw, pw_rest);
			} else {
				fputs(buffer, out_fp);
			}
		} else {
			fputs(buffer, out_fp);
		}
		if (buffer[strlen(buffer) - 1] == '\n') {
			continued = 0;
		} else {
			continued = 1;
		}
		bzero(buffer, sizeof buffer);
	}

	if ((fflush(out_fp)) || (fsync(fileno(out_fp))) || (fclose(out_fp))
		|| (fclose(fp))) {
		unlink(buf);
		return 0;
	}
	if (rename(buf, filename)) {
		return 0;
	} else {
		unlink(lock);
		return 1;
	}
}

static int check_link_count(const char *file)
{
	struct stat sb;

	if (stat(file, &sb) != 0)
		return 0;
	if (sb.st_nlink != 2)
		return 0;
	return 1;
}

static int create_backup(const char *backup, FILE * fp)
{
	struct stat sb;
	struct utimbuf ub;
	FILE *bkfp;
	int c, mask;

	if (fstat(fileno(fp), &sb))
		return -1;

	mask = umask(077);
	bkfp = fopen(backup, "w");
	umask(mask);
	if (!bkfp)
		return -1;

	/* TODO: faster copy, not one-char-at-a-time.  --marekm */
	rewind(fp);
	while ((c = getc(fp)) != EOF) {
		if (putc(c, bkfp) == EOF)
			break;
	}
	if (c != EOF || fflush(bkfp)) {
		fclose(bkfp);
		return -1;
	}
	if (fclose(bkfp))
		return -1;

	ub.actime = sb.st_atime;
	ub.modtime = sb.st_mtime;
	utime(backup, &ub);
	return 0;
}

static int i64c(int i)
{
	if (i <= 0)
		return ('.');
	if (i == 1)
		return ('/');
	if (i >= 2 && i < 12)
		return ('0' - 2 + i);
	if (i >= 12 && i < 38)
		return ('A' - 12 + i);
	if (i >= 38 && i < 63)
		return ('a' - 38 + i);
	return ('z');
}

static char *crypt_make_salt()
{
	time_t now;
	static unsigned long x;
	static char result[3];

	time(&now);
	x += now + getpid() + clock();
	result[0] = i64c(((x >> 18) ^ (x >> 6)) & 077);
	result[1] = i64c(((x >> 12) ^ x) & 077);
	result[2] = '\0';
	return result;
}


static int new_password(const struct passwd *pw, int amroot)
{
	char *clear;
	char *cipher;
	char *cp;
	char orig[200];
	char pass[200];

	if (!amroot && crypt_passwd[0]) {
		if (!(clear = getpass("Old password:"))) {
			return -1;
		}
		cipher = pw_encrypt(clear, crypt_passwd);
		if (strcmp(cipher, crypt_passwd) != 0) {
#ifdef USE_SYSLOG
			syslog(LOG_WARNING, "incorrect password for `%s'",
				   pw->pw_name);
#endif
			sleep(FAIL_DELAY);
			fprintf(stdout, "Incorrect password.\n");
			return -1;
		}
		STRFCPY(orig, clear);
		bzero(clear, strlen(clear));
		bzero(cipher, strlen(cipher));
	} else {
		orig[0] = '\0';
	}
	if (!
		(cp =
		 getpass
		 ("Enter the new password (minimum of 5, maximum of 8 characters)\nPlease use a combination of upper and lower case letters and numbers.\nEnter new password: ")))
	{
		bzero(orig, sizeof orig);
		return -1;
	}
	STRFCPY(pass, cp);
	bzero(cp, strlen(cp));
	if (!obscure(orig, pass, pw)) {
		if (amroot) {
			printf("\nWarning: weak password (continuing).\n");
		} else {
			return -1;
		}
	}
	if (!(cp = getpass("Re-enter new password: "))) {
		bzero(orig, sizeof orig);
		return -1;
	}
	if (strcmp(cp, pass)) {
		fprintf(stdout, "Passwords do not match.\n");
		return -1;
	}
	bzero(cp, strlen(cp));
	bzero(orig, sizeof(orig));
	cp = pw_encrypt(pass, crypt_make_salt());
	bzero(pass, sizeof pass);
	STRFCPY(crypt_passwd, cp);
	return 0;
}

static void set_filesize_limit(int blocks)
{
	struct rlimit rlimit_fsize;

	rlimit_fsize.rlim_cur = rlimit_fsize.rlim_max = 512L * blocks;
	setrlimit(RLIMIT_FSIZE, &rlimit_fsize);
}
