/*
 * usermod - modify users to /etc/passwd and /etc/shadow
 *
 *
 * Copyright (C) 1999 by Moreton Bay LLC.
 * Written by Matthew Natalier <matthewn@moreton.com.au>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
#include "internal.h"
#include "shadow_.h"
#include "gshadow_.h"
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <getopt.h>
#include <grp.h>
#include "group.h"
//#include "../../linux/include/linux/autoconf.h"

/*#define GRP_FILE "/etc/config/group"*/

struct groupcheck{
	char *name;
	int added;
};

#define GRP_ADMIN "root" /*always make root the group admin*/

#define FLATFSD_PID_FILE ("/var/log/flatfsd.pid")

/*
#define GSHADOW_FILE "etc/config/gshadow"
*/


#ifndef CONFIG_REDUCED_MEMORY
static const char usermod_usage[] =
	"usermod [OPTION] login\n\n"
	"Modify a user parameters\n"
	"-g initial_group - the group that the"
	"user should be in.\n"
	"-G grp1,grp2,... - comma seperated list of other groups.\n";
//	"-u uid - the new user id of the user.\n"
#else
static const char usermod_usage [] = 
	"usermod [OPTION] login\n\n";
#endif //CONFIG_REDUCED_MEMORY

#define FLAGS "g:G:c:"//u:"

//
//Debug stuff!!
//
#define DBG_USER_MOD 0

#if DBG_USER_MOD
#define dbgprint(x) printf x
#else
#define dbgprint(x)
#endif //DBG_USER_MOD


int
AddUserToGroup(struct group *gr, char *usr);

void
RemoveUserFromGroup(struct group *grp, char *userName);

int
UserInGroup(struct group *grp, char *user);

int usermod_main(int argc, char **argv)
{
	int Gflg = 0;			//-G supplementary groups.
	int gflg = 0;			//-g the main group.
//	int uflg = 0;			//-u new uid for the user.
	int cflg = 0;			//-c set the comment field for the user.

	char *def_group = NULL;		//the new default group for the user
	char *sup_groups = NULL;	//other groups that the user should becom part of.
//	char *uid;				//the new user id that the user should become.

	struct group * grp;
	struct passwd *pw;
	char flag;
	char *user;
	char *gecos;

	//MN USAGE
	//This is probably not correct, but we don't let user change there own
	//configuration.
	//
//#if !(DBG_USER_MOD)
	if(getuid() != 0)
	{
		fprintf(stderr, "You must be root to use this function\n");
		exit(1);
	}
//#endif

	while ((flag = getopt(argc, argv, FLAGS)) != EOF) {
		switch (flag) {
		case 'G':
			sup_groups = optarg;
			dbgprint(("USERMOD: Specified other groups\n"));
			Gflg++;
			break;
		case 'g':
			def_group = optarg;
			dbgprint(("USERMOD: Specified new major group\n"));
			gflg++;
			break;
/*		case 'u':
			uid = optarg;
			uflg++;
			break;*/
		case 'c':
			gecos = optarg;
			cflg++;
			break;
		default:
			usage(usermod_usage);
		}
	}
	user = argv[optind];

	/*
	We probably don't need to do this crap.
	*/
#if 0
	if((!gflg) && (!Gflg)){				/*They aren't modifying anything*/
//		fprintf(stderr,"usermod: Nothing to modify.\n");
		usage(usermod_usage);
		exit(1);
	}
	if((gflg) && (!Gflg)){			/*They are only offering the main group*/
		UpdateGrpInformation(user, NULL, def_group);
	}
	else if((!gflg) && (Gflg)){			/*They are only modifying supplemental groups*/
		UpdateGrpInformation(user, sup_groups, NULL);
	}
	else if((gflg) && (Gflg)){						/*They are modifying both grps*/
#endif
	UpdateGrpInformation(user, sup_groups, def_group);

	if(cflg){
		UpdateUserInformation(user, gecos);
	}

	exit(0);
}


int
GroupInList(char *list, char *grp)
/*
desc:	Does grp exist in a comma delimited list.
*/
{
	char *tmp;
	char *nxt;
	/*can't use strtok because we don't want to change the contents*/
	tmp = list;
	do{
		nxt = strstr(list, grp);
		if(nxt == NULL)
			return 0;
		if((nxt == tmp) || ((nxt - 1)[0] == ',')) {
			if(strlen(grp) > strlen(nxt))
				return 0;
			else if((nxt[strlen(grp)] == '\0') || (nxt[strlen(grp)] == ','))
				return 1;
		}
		tmp = nxt;
		nxt = strstr(nxt, grp);
	}while(nxt != NULL);
	return 0;
}



int
UpdateGrpInformation(char *usrName, char *sup_grps, char *main_grp)
/*
desc:	If we don't have a main_grp then we don't change this inormation
		otherwise we change this information
TO DO:	Modify shadow group information.
*/
{
	struct passwd *usr;
	struct group *grp;
	FILE *in;
	FILE *out;
	int addMain = 0;

	if(main_grp != NULL){
		dbgprint(("USERMOD: need to modify default group\n"));
		usr = getpwnam(usrName);
		if(!usr){
			fprintf(stderr,"usermod: No such user name.");
			return 1;
		}
		grp = getgrnam(main_grp);
		addMain = 1;
		if(grp->gr_gid != usr->pw_gid){
		/*TO DO!!
		We need to actually change the main group for the user at this point!!
		do we need to do anything related to the shadow stuff??
		*/
		}
	}
	in = fopen(GRP_FILE,"r");
	if(!in){
		fprintf(stderr,"usermod: Cannot open group file - %s\n", strerror(errno));
		return 1;
	}
	out = fopen("/etc/config/tmpfile2", "w");
	if(!out){
		fprintf(stderr,"usermod: Unable to open temporary file - %s\n" , strerror(errno));
		return 1;
	}
#ifdef TLG_FEATURE_SHADOWPASSWDS
	/*Open shadow password files in here.*/
#endif /*TLG_FEATURE_SHADOWPASSWDS*/
/*
	shin = fopen(SGROUP_FILE, "r");
	if(shin){
		fprintf(stderr, "usermod: Canoot open shadow groups - %s\n", */
	grp = fgetgrent(in);
	while(grp != NULL){
		if(sup_grps && GroupInList(sup_grps, grp->gr_name)){/*the user needs to be added to this group*/
			dbgprint(("USERMOD: Adding user to group, %s\n",grp->gr_name));
			gr_addusr(grp, usrName);
#ifdef TLG_FEATURE_SHADOWPASSWDS
			/*TO DO - update shadow stuff*/
#endif /*TLG_FEATURE_SHADOWPASSWDS*/
		}
		else if(addMain && (strcmp(main_grp, grp->gr_name) == 0)){/*The user needs to be added to this group*/
			dbgprint(("USERMOD: Adding user to main group - %s\n", grp->gr_name));
			gr_addusr(grp, usrName);
#ifdef TLG_FEATURE_SHADOWPASSWDS
			/*TO DO update shadow stuff*/
#endif /*TLG_FEATURE_SHADOWPASSWDS*/
		}
		/*It may be better to always try and remove the user in an else conditions??*/
		else if(UserInGroup(grp, usrName)){/*The user needs to be removed from this group*/
			dbgprint(("USERMOD: Removing a user from a group - %s\n", grp->gr_name));
			RemoveUserFromGroup(grp, usrName);
#ifdef TLG_FEATURE_SHADOWPASSWDS
			/*TO DO update shadow stuff*/
#endif /*TLG_FEATURE_SHADOWPASSWDS*/
		}
		else;/*This group doesn't need to be modified*/
		writegr(out, grp);
		grp = fgetgrent(in);
	}
	fclose(out);
	fclose(in);

#ifdef TLG_FEATURE_SHADOWPASSWDS
	/*TO DO - close the files in here*/
#endif /*TLG_FEATURE_SHADOWPASSWDS*/
	
	unlink(GRP_FILE);
	link("/etc/config/tmpfile2", GRP_FILE);
	unlink("/etc/config/tmpfile2");

	/*TO DO!!
	Write information to flash
	*/
//	commitChanges();
}

void
RemoveUserFromGroup(struct group *grp, char *userName)
/*
desc:	The user is removed from this group.
*/
{
	int i;
	int j;
	if((grp == NULL) || (userName == NULL)){
		dbgprint(("Null parameters\n"));
		return;
	}

	dbgprint(("Remove_UserFromGroup: entering, user - %s, grp - %s\n", userName, grp->gr_name));

	for(i=0,j=0;grp->gr_mem[j] != NULL;i++,j++){
		if(strcmp(grp->gr_mem[j], userName) == 0)
			j++;
		grp->gr_mem[i] = grp->gr_mem[j];
	}
	grp->gr_mem[i] = NULL;
}

int
UserInGroup(struct group *grp, char *user)
/*
desc:	This checks to see if a user is present in the group.
*/
{
	int i;
	for(i=0;grp->gr_mem[i] != NULL;i++){
		if(strcmp(grp->gr_mem[i], user) == 0){
			dbgprint(("UserInGroup: group name - %s, user - %s\n", grp->gr_name,user));
			return 1;
		}
	}
	return 0;
}

/*
 * commitChanges
 *
 * Burns all files into the flash
 *
 * retn:	0 on success, -1 on error
 */
int commitChanges() {
	pid_t pid;
	FILE *in;
		
	/* get the pid of flatfsd */
	in = fopen(FLATFSD_PID_FILE, "r");
	
	if(in == NULL) {
		/* couldn't access flatfsd pid file */
#if DATABASE_DEBUG
		printf("Couldn't find FLATFSD_PID_FILE<BR>", pid);
#endif
		return -1;
	}

	if(fread(&pid, sizeof(pid_t), 1, in) > 0) {
		/* we read something.. hopefully the pid */
#if DATABASE_DEBUG
		printf("pid = %d<BR>", pid);
#endif
	} else {
		/* no data read from file */
#if DATABASE_DEBUG
		printf("No data read from file<BR>", pid);
#endif
		fclose(in);
		return -1;
	}
	fclose(in);
	
	/* send that pid signal 10 */
	if((kill(pid, 10)) == -1) {
#if DATABASE_DEBUG
		printf("oops! couldn't send signal 10 to pid %d<BR>", pid);
		perror("kill failed");
#endif
		return -1;
	}
#if DATABASE_DEBUG
	printf("file changes commited<BR>");
#endif
	return 0;
}

int
UpdateUserInformation(char *username, char *gecos)
{
	FILE *in;
	FILE *out;
	struct passwd *pwd;
	
	in = fopen(PASSWD_FILE, "r");
	if(!in)
		return 0;
	out = fopen("/etc/config/tmp", "w");
	if(!out)
		return 0;

	while((pwd = fgetpwent(in))  != NULL){
		if(strcmp(pwd->pw_name, username) == 0){
			pwd->pw_gecos = gecos;
		}
		putpwent(pwd, out);
	}

	fclose(in);
	fclose(out);

	unlink(PASSWD_FILE);
	link("/etc/config/tmp", PASSWD_FILE);
	unlink("etc/config/tmp");
}



